-- enemy.lua
local colors = require 'modules.setup.colors'
local settings = require 'modules.setup.settings'
local utils = require 'modules.setup.utils'
local floatingText = require 'modules.ui.floatingtext'
local enemylist = require 'modules.enemies.enemylist'
local item = require 'modules.items.item'

local stateslist = require 'modules.enemies.stateslist'


local defaults = {
  fg = colors.white,
  bg = colors.black,
  name = 'Enemy',
  enemytype = 'default',
  quadname = 'enemy',
  visible = false,
  blocksMovement = true,
  blocksVision = false,
  isEnemy = true,
  zIndex = 5,
  hp = 10,
  maxhp = 10,
  state = 'spawning',
  status = nil,
  ac = 10
}

local enemy = Class{__includes = {
  defaults
}}

function enemy:init(x, y, args)
  local ts = settings.tilesize
  self.x = x or 1
  self.y = y or 1
  self.drawX = (self.x-1) * ts
  self.drawY = (self.y-1) * ts
  self.w = ts
  self.h = ts
  self.tweenSpeed = .1

  utils.copy(self, args or {})
  utils.copy(self, enemylist[self.enemytype])

  if self.state == 'spawning' then
    local hd = enemylist[self.enemytype].hd
    self.maxhp = utils.roll(hd[1],hd[2],hd[3])
    self.hp = self.maxhp
    self:chooseInitialState()
  end

  self.hitdice = enemylist[self.enemytype].hd[1]

  if self.postInit then self:postInit() end
end

function enemy:draw()
  if self.visible then
    local ts = settings.tilesize
    local x, y, w, h = self.drawX, self.drawY, self.w, self.h
    local quad
    if stateslist[self.state].quadname and os.time() % 2 == 0 then
      quad = stateslist[self.state].quadname
    else
      quad = self.quadname
    end
    lg.setColor(self.bg)
    lg.rectangle('fill', x, y, w, h)
    lg.setColor(self.fg)
    lg.draw(QualeTiles, QualeQuads[quad], x, y)
    lg.setColor(colors.white)
  end
end

function enemy:updateVisibility(vis)
  local x,y = self.x, self.y
  if vis[y] and vis[y][x] == 1 then
    self.visible = true
  else
    self.visible = false
  end
end

function enemy:receiveDamage(amt, source, isCrit)
  local isCrit = isCrit or false
  -- eventually, armor and damage reduction will occur here:
  -- waking up:
  if self.state == 'sleep' then
    self.state = 'moveTowardsPlayer'
  end
  -- damage:
  self.hp = self.hp - amt
  local msg
  if isCrit then
    msg = 'Critical hit! ' .. amt .. ' damage!'
  else
    msg = 'The ' .. self.name .. ' took ' .. amt .. ' damage.'
  end
  Signal.emit('message', msg)
  -- floating text:
  local fstring = amt
  if isCrit then fstring = fstring .. '!' end
  Signal.emit('addEntity', floatingText({
    text = fstring,
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize
  }))
  if self.hp <= 0 then self:kill(source) end
end

function enemy:kill(source)
  local xp = self.hitdice or 1
  local msg = 'You killed the ' .. self.name .. '! '
  msg = msg .. 'Gained ' .. xp .. ' XP.'
  Signal.emit('message', msg)
  if self.droplist then
    self:dropItems()
  end
  if source.gainXP then source:gainXP(xp) end
  self._remove = true
end

function enemy:save()
  return {
    type = 'Enemy',
    x = self.x,
    y = self.y,
    args = {
      hp = self.hp,
      enemytype = self.enemytype,
      state = self.state,
      status = self.status
    }
  }
end

function enemy:dropItems()
  local x,y = self.x, self.y
  local list = enemylist[self.enemytype].droplist
  local type = ROT.RNG:getWeightedValue(list)
  Signal.emit('addEntity', item(x, y, { itemtype = type, visible = self.visible }))
end

-----------------------------------------------------------
-- movement
-----------------------------------------------------------
function enemy:moveTo(x, y)
  self.x, self.y = x, y
  Timer.tween(
    self.tweenSpeed,
    self,
    {
      drawX = (x-1) * settings.tilesize,
      drawY = (y-1) * settings.tilesize
    },
    'out-quad',
    function()
      self.drawX = (x-1) * settings.tilesize
      self.drawY = (y-1) * settings.tilesize
    end)
end

function enemy:blockedPath(x, y, entities, len)
  for i=1,len do
    local ent = entities[1]
    if ent.isPlayer then
      self:attack(ent)
      break
    end
  end
end

function enemy:attack(target)
  local n = self.name or 'foe'
  local str = nil
  -- todo: actual combat stuff here.
  if target.drawX and target.drawY then
    self:meleeAttackAnimation(target.drawX, target.drawY)
  end
  local tohit = enemylist[self.enemytype].tohit

  local roll = utils.roll(tohit)

  if roll == 20 then
    target:receiveDamage(self:getCriticalDamage(), self, true)
    str = 'The ' .. n .. ' critically strikes!'
  elseif roll >= target.ac then
    target:receiveDamage(self:getDamage(), self, false)
    str = 'The ' .. n .. ' attacks you.'
  else
    self:miss(target)
  end

  if str then Signal.emit('message', str) end

  target:receiveDamage(self:getDamage(), self)
end

function enemy:miss()
  Signal.emit('addEntity', floatingText({
    text = 'Miss',
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize
  }))
end

function enemy:getDamage()
  local damage = enemylist[self.enemytype].damage

  return utils.rollTable(damage)
end

function enemy:getCriticalDamage()
  local dtable = enemylist[self.enemytype].damage

  damage = dtable[1] * dtable[2] + dtable[3]
  return damage
end

function enemy:meleeAttackAnimation(drawX, drawY)
  local dx,dy = drawX, drawY
  local sdx, sdy = self.drawX, self.drawY
  local speed = self.tweenSpeed / 2
  Timer.tween(speed, self,{drawX = dx, drawY = dy },'out-quad',
    function() Timer.tween( speed, self, {drawX = sdx, drawY = sdy}, 'out-quad', function()
          self.drawX = sdx
          self.drawY = sdy
        end)
    end)
end

function enemy:getDamage()
  -- temp:
  return love.math.random(1,3)
end

-----------------------------------------------------------
-- Acting and behaviors
-----------------------------------------------------------

function enemy:act()
  self:chooseState()
  local s = stateslist[self.state]
  s.action(self)
end

function enemy:chooseState()
  -- if low health, move randomly because we don't have a flee state
  if self.hp <= self.maxhp * .2 then
    self.state = 'moveRandom'
  elseif self.state == 'sleep' and self:distanceToPlayer() <= 3 then
    local wakeupChance = 0.5
    if love.math.random() < wakeupChance then
      self.state = 'moveTowardsPlayer'
    end
  elseif self.state ~= sleep and self.visible == false then
    local sleepChance = 0.5
    if love.math.random() < sleepChance then
      self.state = 'sleep'
    end
  end
end

function enemy:chooseInitialState()

  if self.immuneToSleep then
    self.state = 'moveRandom'
  else
    local sleepChance = 0.5
    if love.math.random() < sleepChance then
      self.state = 'sleep'
    else
      self.state = 'moveRandom'
    end
  end
end

function enemy:distanceToPlayer()
  local x1,y1 = self.x, self.y
  local x2,y2 = Game.player.x, Game.player.y
  return ((x2-x1)^2+(y2-y1)^2)^0.5
end

return enemy
