local colors = require 'modules.setup.colors'
local settings = require 'modules.setup.settings'
local utils = require 'modules.setup.utils'
local itemlist = require 'modules.items.itemlist'
local itemmixins = require 'modules.items.itemmixins'
local levellist = require 'modules.level.levellist'

local defaults = {
  visible = false,
  blocksMovement = false,
  blocksVision = false,
  isEnemy = false,
  isItem = true,
  zIndex = 4,
  quadname = 'unknown',
  name = 'Item',
  type = 'Item',
  itemtype = 'nothing',
  autopickup = false,
  isEquipment = false,
  isConsumable = false,
  isStackable = false,
  isFood = false
}

local item = Class{__includes = {
  require 'modules.tile',
  defaults
}}
-----------------------------------------------------------
-- init
-----------------------------------------------------------
function item:init(x,y,args)
  local ts = settings.tilesize

  self.x = x or 1
  self.y = y or 1
  self.w = ts
  self.h = ts

  self.drawX = (self.x-1) * ts
  self.drawY = (self.y-1) * ts

  self.fg = colors.gray
  self.bg = colors.black

  -- if depth, give self an itemlevel:
  if Game.depth and levellist[Game.depth] then
    local min = itemlist[self.itemtype].minItemLevel
    local max = itemlist[self.itemtype].maxItemLevel
    self.itemlevel = love.math.random(min, max)
  end

  utils.copy(self, args or {})
  utils.copy(self, itemlist[self.itemtype])

  if itemlist[self.itemtype].stats then self:doStats() end
  if self.postInit then self:postInit() end

  -- temp:
  local magic = love.math.random() < .9--levellist[Game.depth].magicItemProbability
  if magic and self.isFood then
    local p = table.random(itemmixins.foodprefixes)
    self.name = p .. ' ' .. self.name
    self.bonus = love.math.random(1, Game.depth)
    self.bonustype = 'Healing'
  elseif magic then
    local magictype = ROT.RNG:getWeightedValue({
      prefix = 40,
      suffix = 40,
      both = 20
    })
    if magictype == 'prefix' then
      local p = table.random(itemmixins.prefixes)
      self.name = p.name .. ' ' .. self.name
      self.bonustype = p.type
      self.bonus = love.math.random(1, Game.depth)
    elseif magictype == 'suffix' then
      local s = table.random(itemmixins.suffixes)
      self.name = self.name .. ' ' .. s.name
      self.bonustype = s.type
      self.bonus = love.math.random(1, Game.depth)
    else
      local p = table.random(itemmixins.prefixes)
      local s = table.random(itemmixins.suffixes)
      self.name = p.name .. ' ' .. self.name .. ' ' .. s.name
      self.bonustype = s.type
      self.bonus = love.math.random(1, Game.depth) * 2
    end
  end

end

-----------------------------------------------------------
-- stats functions
-----------------------------------------------------------
function item:doStats()
  local stats = itemlist[self.itemtype].stats
  local itemlevel = self.itemlevel
  self.stats = {}

  for k,v in pairs(stats) do
    self.stats[k] = love.math.random(stats[k].min, stats[k].max)
  end
end

-----------------------------------------------------------
-- drawing functions
-----------------------------------------------------------
function item:draw()
  if self.visible then
    local ts = settings.tilesize
    local x, y, w, h = self.drawX, self.drawY, self.w, self.h
    local quad = QualeQuads[self.quadname]
    lg.setColor(self.bg)
    lg.rectangle('fill', x, y, w, h)
    lg.setColor(self.fg)
    lg.draw(QualeTiles, quad, x, y)
    lg.setColor(colors.white)
  end
end

function item:updateVisibility(vis)
  local x,y = self.x, self.y
  if vis[y] and vis[y][x] == 1 then
    self.visible = true
  else
    self.visible = false
  end
end

-----------------------------------------------------------
-- save functions
-----------------------------------------------------------
function item:save()
  return {
    type = 'Item',
    x = self.x,
    y = self.y,
    args = {
      name = self.name,
      itemtype = self.itemtype,
      itemlevel = self.itemlevel,
      value = self.value or nil,
      damage = self.damage or nil,
      ac = self.ac or nil,
      bonus = self.bonus or nil,
      bonustype = self.bonustype or nil
    }
  }
end

function item:getInventoryInfo()
  local d = {
    name = self.name,
    itemlevel = self.itemlevel,
    itemtype = self.itemtype,
    quadname = self.quadname,
    damage = self.damage or nil,
    value = self.value or nil,
    ac = self.ac or nil,
    bonus = self.bonus or nil,
    bonustype = self.bonustype or nil
  }
  local i = self.inventoryInfo or {}
  utils.copy(d, i)
  return d
end

return item
