local settings = require 'modules.setup.settings'
local colors = require 'modules.setup.colors'
local map = require 'modules.level.map'
local levellist = require 'modules.level.levellist'

local enemy = require 'modules.enemies.enemy'
local mapfeature = require 'modules.level.mapfeature'
local item = require 'modules.items.item'
local itemlist = require 'modules.items.itemlist'

local levelgen = Class{}
-----------------------------------------------------------
-- Init
-----------------------------------------------------------
function levelgen:init()

end

-----------------------------------------------------------
-- MakeLevel functions
-----------------------------------------------------------

function levelgen:makeLevel(depth)
  local map = {}
  local entities = {}
  local config = levellist[depth]

  map = self:makeMap(config)
  entities = self:addEntities(config, map)

  return map, entities
end

-----------------------------------------------------------
-- Map functions
-----------------------------------------------------------
function levelgen:makeMap(config)
  return map(config.mapW, config.mapH)
end
-----------------------------------------------------------
-- Entity functions
-----------------------------------------------------------
function levelgen:addEntities(config, map)
  local entities = {}
  entities = self:addMapfeatures(config, entities, map)
  entities = self:addEnemies(config, entities, map)
  entities = self:addItems(config, entities, map)

  return entities
end

function levelgen:addMapfeatures(config, entities, map)
  local c = config
  local ents = entities
  local map = map
  local rooms = map.rooms

  -- place stairs
  local r1 = table.random(rooms)
  local r2
  repeat
    r2 = table.random(rooms)
  until r2 ~= r1

  -- stairs up
  local x = math.random(r1._x1, r1._x2)
  local y = math.random(r1._y1, r1._y2)
  table.insert(ents, mapfeature(x, y, {mapfeaturetype = 'stairsup'}))

  -- stairs down
  x = math.random(r2._x1, r2._x2)
  y = math.random(r2._y1, r2._y2)
  table.insert(ents, mapfeature(x, y, {mapfeaturetype = 'stairsdown'}))

  -- place doors:
  for i=1, #rooms do
    local room = rooms[i]
    local doors = room._doors
    for k, v in pairs(doors) do
      if love.math.random() < config.doorProbability then
        -- place door:
        local location = k
        local parts = location:split(',')
        local x,y = tonumber(parts[1]), tonumber(parts[2])
        -- don't place doors on top of each other:
        if self:spaceIsFree(x, y, map, ents) then
          local isOpen = table.random({true,false})
          table.insert(ents, mapfeature(x, y, {
            mapfeaturetype='door',
            open = isOpen
          }))
        end
      end
    end
  end

  -- place flowers:
  for i=1,10 do
    local x, y = self:getFreeSpace(map, ents)
    table.insert(ents, mapfeature(x,y,{mapfeaturetype='tinyflower'}))

    x, y = self:getFreeSpace(map, ents)
    table.insert(ents, mapfeature(x,y,{mapfeaturetype='flower'}))

  end


  return ents
end

function levelgen:addEnemies(config, entities, map)
  local c = config
  local ents = entities
  local map = map
  local count = love.math.random(c.minEnemies, c.maxEnemies)

  for i=1, count do
    local x, y = self:getFreeSpace(map, ents)
    local list = config.enemyProbabilities
    local type = ROT.RNG:getWeightedValue(list)
    local e = enemy(x, y, {enemytype = type})
    table.insert(ents, e)
  end

  return ents
end

function levelgen:addItems(config, entities, map)
  local c = config
  local ents = entities
  local map = map
  local count = love.math.random(c.minItems, c.maxItems)

  for i=1, count do
    local x, y = self:getFreeSpace(map, ents)
    local list = config.itemProbabilities
    local type = ROT.RNG:getWeightedValue(list)
    local e = item(x, y, {itemtype = type})
    table.insert(ents, e)
  end

  return ents
end


-----------------------------------------------------------
-- Utility/helper functions
-----------------------------------------------------------
function levelgen:getFreeSpace(map, entities)
  local tiles = map.tiles
  local ex, ey
  repeat
    ex, ey = love.math.random(1, #tiles[1]), love.math.random(1, #tiles)
  until self:spaceIsFree(ex,ey, map, entities)
  return ex, ey
end

function levelgen:spaceIsFree(x, y, map, entities)
  local tiles = map.tiles
  local entities, len = self:entitiesAt(x,y, entities)
  if (tiles[y] and tiles[y][x] == 0) and (len == 0) then
    return true
  else
    return false
  end
end

function levelgen:entitiesAt(x,y, entities, filter)
  local list = {}
  local f = filter or function(e) return true end

  for _,item in ipairs(entities) do
    if item.x == x and item.y == y and f(item) then
      table.insert(list, item)
    end
  end
  return list, #list
end

return levelgen
