-- map.lua
local settings = require 'modules.setup.settings'
local colors = require 'modules.setup.colors'
local map = Class{}

-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function map:init(a,b)
--[[
  if first argument is a table, it's a tile map, so load it.
  if first and second args are numbers, it's dimensions for
  a new map, so create a new map.
--]]
  if type(a) == 'table' then
    self:load(a)
  else
    local w = self.w or a
    local h = self.h or b
    self:newMap(w, h)
  end
  self:setUpCanvas()
end

function map:setUpCanvas()
  local ts = settings.tilesize
  local cw, ch = self.w * ts, self.h * ts
  self.cache = lg.newCanvas(cw, ch)
end

function map:newMap(w, h)
  self.w = w
  self.h = h

  self.tiles = self:makeTiles(w, h, 1)
  self.tiles = self:makeDungeon(self.tiles)

  self.seen = self:makeTiles(w, h, 0)
  self.visible = self:makeTiles(w, h, 0)
end

function map:makeTiles(w, h, fill)
  local tiles = {}
  local fillvalue = fill or 0
  for y=1,h do
    tiles[y] = {}
    for x=1,w do
      tiles[y][x] = fillvalue
    end
  end
  return tiles
end

function map:makeDungeon(t)
  local w,h = self.w, self.h
  local tiles = t
  local m = ROT.Map.Uniform:new(w,h)
  local callback = function(x, y, val)
    tiles[y][x] = val
  end
  m:create(callback)
  self.rooms = m:getRooms()
  self.doors = m:getDoors()
  return tiles
end

-----------------------------------------------------------
-- draw functions
-----------------------------------------------------------
function map:draw()
  lg.draw(self.cache)
end

function map:drawToCache()
  lg.setCanvas(self.cache)
  local ts, tiles = settings.tilesize, self.tiles
  for y=1,#tiles do
    for x=1,#tiles[1] do
      local tile = self:getTile(x,y)
      local seen = self:getSeen(x,y)
      local visible = self:getVisible(x,y)

      local dx, dy, w, h = (x-1) * ts, (y-1) * ts, ts, ts
      local bg, fg, quad

      if seen or visible then
        -- choose quad
        if tile == 1 then
          quad = 'hash'
          bg = 'black'
          fg = 'lightblue'
        elseif tile == 0 then
          quad = 'dot'
          bg = 'black'
          fg = 'darkblue'
        end

        -- choose seen-only color
        if seen and not visible then
          fg = 'purple'
          bg = 'black'
        end
        self:drawTile(dx, dy, quad, fg, bg, w, h)
      end

    end
  end
  lg.setCanvas()
end

function map:drawTile(dx, dy, quad, fg, bg, w, h)
  lg.setColor(colors[bg])
  lg.rectangle('fill', dx, dy, w, h)
  lg.setColor(colors[fg])
  lg.draw(QualeTiles, QualeQuads[quad], dx, dy)
  lg.setColor(colors.white)
end

-----------------------------------------------------------
-- Helper functions
-----------------------------------------------------------
function map:getDimensions()
  return self.w, self.h
end

function map:getTile(x, y)
  return self.tiles[y][x]
end

function map:getSeen(x, y)
  return self.seen[y][x] == 1
end

function map:getVisible(x, y)
  return self.visible[y][x] == 1
end

-----------------------------------------------------------
-- FOV
-----------------------------------------------------------
function map:doFOV(px, py, blockingEntities)
  local radius, startA, endA, perm = 10, 0, math.pi * 2, 10
  -- reset visibility map:
  self.visible = self:makeTiles(self.w, self.h, false)

  for i=1,#blockingEntities do
    local entity = blockingEntities[i]
    local x, y = entity.x, entity.y
    -- TODO
  end

  -- functions:
  local function isTransparent(x, y)
    -- return true if the cell is non-blocking
    if self.tiles[y] and self:getTile(x,y) == 0 then
      return true
    else
      return false
    end
  end

  local function onVisible(x, y)
  	-- gets called when a square is visible
    local dx,dy = x-px,y-py
    -- circular radius:
		if (dx*dx + dy*dy) > radius*radius + radius then
			return
		end
    if self.tiles[y] then
      self.seen[y][x] = 1
      self.visible[y][x] = 1
    end
  end

  FOV(px, py, radius, isTransparent, onVisible, startA, endA, perm)
  self:drawToCache()
end

-----------------------------------------------------------
-- Saving and loading
-----------------------------------------------------------
function map:save()
  -- returning these and then saving them uncompressed is
  -- SLIGHTLY inefficient, but it makes for a readable save file, so...
  -- yeah.
  return {
    tiles = self.tiles,
    seen = self.seen,
  }
end

function map:load(savedata)
  self.tiles = savedata.tiles
  self.seen = savedata.seen

  self.w = #self.tiles[1]
  self.h = #self.tiles

  -- note: visible is not saved or loaded.
  -- Make sure to do a doFOV(player.x, player.y) after loading.
end

return map
