-- player.lua
local colors = require 'modules.setup.colors'
local settings = require 'modules.setup.settings'
local utils = require 'modules.setup.utils'
local floatingText = require 'modules.ui.floatingtext'
local itemlist = require 'modules.items.itemlist'
local statuslist = require 'modules.status.statuslist'

local player = Class{__includes = {

}}

-----------------------------------------------------------
-- Init
-----------------------------------------------------------
function player:init(x, y, args)
  local ts = settings.tilesize
  self._dead = false
  self.x = x or 1
  self.y = y or 1
  self.drawX = (self.x-1) * ts
  self.drawY = (self.y-1) * ts
  self.w = ts
  self.h = ts
  self.quadname = table.random({'man','woman'})
  self.fg = colors.white
  self.bg = colors.black
  self.tweenSpeed = .1
  self.name = 'Player'
  self.isPlayer = true
  self.visible = true
  self.blocksVision = false
  self.zIndex = 10
  self.maxInventorySize = 12

  self.statuses = {}

  self.equipment = {
    head = {
      name = 'Head'
    },
    body = {
      name = 'Body'
    },
    hand = {
      name = 'Hand'
    },
    offhand = {
      name = 'Hand'
    }
  }
  self.inventory = {}

  utils.copy(self, args or {
    hp = 20,
    maxhp = 20,
    ac = 0,
    gold = 0,
    keys = 1,
    level = 1,
    xp = 0,
    mind = self:getStatRoll(),
    str = self:getStatRoll(),
    dex = self:getStatRoll()
  })

  if args == nil then
    self.maxhp = self.str + 6
    self.hp = self.maxhp
  end

  self:calculateStats()

  Signal.register('playerFinishMovement', function()
    self:finishMovement()
  end)
end

function player:getStatRoll()
  -- returns a value between 16 and 18
  -- because the game's too hard with actually-random stats
  return love.math.random(16,18)
end

-----------------------------------------------------------
-- Movement
-----------------------------------------------------------
function player:moveTo(x, y)
  Signal.emit('lock')
  self.x, self.y = x, y
  Signal.emit('updateFOV', x, y)
  Timer.tween(
    self.tweenSpeed,
    self,
    {
      drawX = (x-1) * settings.tilesize,
      drawY = (y-1) * settings.tilesize
    },
    'out-quad',
    function()
      self.drawX = (x-1) * settings.tilesize
      self.drawY = (y-1) * settings.tilesize
      Signal.emit('unlock')
      Signal.emit('playerFinishMovement')
    end)
end

function player:teleportTo(x,y)
  Signal.emit('lock')
  self.x, self.y = x, y
  self.drawX = (x-1) * settings.tilesize
  self.drawY = (y-1) * settings.tilesize
  Signal.emit('updateFOV', x, y)
  Signal.emit('unlock')
  Signal.emit('playerFinishMovement')
end

function player:blockedPath(x, y, entities, len)
  Signal.emit('lock')
  local str = 'An object blocks your path.'
  --handle path blocked by entities
  for i=1,len do
    local ent = entities[1]
    if ent.isEnemy then
      self:attack(ent)
      break
    elseif ent.name == 'Door' then
      self:openDoor(ent)
    elseif ent.blocksMovement then
      if ent.name then
        str = 'A ' .. ent.name .. ' blocks your path.'
        -- unlocking occurs in attack anim function
      end
      Signal.emit('message', str)
      break
    end
  end
  Signal.emit('unlock')
end

-----------------------------------------------------------
-- Drawing
-----------------------------------------------------------
function player:draw()
  local ts = settings.tilesize
  local x, y, w, h = self.drawX, self.drawY, self.w, self.h
  lg.setColor(self.bg)
  lg.rectangle('fill', x, y, w, h)
  lg.setColor(self.fg)
  lg.draw(QualeTiles, QualeQuads[self.quadname], x, y)
  lg.setColor(colors.white)
end

-----------------------------------------------------------
-- Gameplay
-----------------------------------------------------------
function player:receiveDamage(amt, source, isCrit)
  -- eventually, armor and damage reduction will occur here:
  self.hp = self.hp - amt
  local name = source.name or 'foe'
  local msg = 'The ${n} dealt ${a} damage!' % { n = name, a = amt}
  Signal.emit('message', msg)

  -- spawn floating text

  Signal.emit('addEntity', floatingText({
    text = amt,
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize
  }))
  Signal.emit('screenshake')
  if self.hp <= 0 then
    self.hp = 0
    self:kill()
  end

end

function player:heal(amt, source)
  if self.hp == self.maxhp then
    local msg =  'Your health is already full.'
    Signal.emit('message', msg)
    return false
  else
    self.hp = math.min(self.hp + amt, self.maxhp)
    local msg =  'You healed ${a} HP.' % {a = amt}
    Signal.emit('message', msg)
    Signal.emit('addEntity', floatingText({
      text = '+' .. amt,
      x = self.drawX + self.w / 2,
      y = self.drawY - settings.tilesize,
      color = colors.lightgreen
    }))
    return true
  end
end

function player:kill()
  self.fg = colors.red
  self._dead = true
  Signal.emit('message', 'You died... [Enter] to restart')
  Signal.clear('playerFinishMovement')
end

function player:getDamage()
  local damage
  if self.equipment.hand.damage then
    damage = utils.rollTable(self.equipment.hand.damage)
  else
    damage = utils.roll(1,3,0)
  end
  damage = damage + self.damagebonus
  return damage
end

function player:getCriticalDamage()
  local damage
  if self.equipment.hand.damage then
    local dtable = self.equipment.hand.damage
    damage = dtable[1] * dtable[2] + dtable[3]
  else
    damage = 3
  end
  return damage + self.damagebonus
end

function player:attack(enemy)
  -- todo: actual combat stuff here.
  if enemy.drawX and enemy.drawY then
    self:meleeAttackAnimation(enemy.drawX, enemy.drawY)
  end

  local roll = utils.roll(1, 20, 0)

  if roll == 20 then
    enemy:receiveDamage(self:getCriticalDamage(), self, true)
  elseif roll + self.tohit >= enemy.ac then
    enemy:receiveDamage(self:getDamage(), self, false)
  else
    self:miss(enemy)
  end
end

function player:miss(enemy)
  local n = enemy.name or 'foe'
  Signal.emit('message', 'You miss the ' .. n .. '.')
  Signal.emit('addEntity', floatingText({
    text = 'Miss',
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize
  }))
end

function player:meleeAttackAnimation(drawX, drawY)
  local dx,dy = drawX, drawY
  local sdx, sdy = self.drawX, self.drawY
  local speed = self.tweenSpeed / 2
  Timer.tween(speed, self,{drawX = dx, drawY = dy },'out-quad',
    function() Timer.tween( speed, self, {drawX = sdx, drawY = sdy}, 'out-quad', function()
          self.drawX = sdx
          self.drawY = sdy
          Signal.emit('unlock')
        end)
    end)
end

function player:finishMovement()
  local x, y = self.x, self.y

  -- collect autopickup items at this location:
  local filter = function(e) return e.autopickup == true end
  local items, len = Game:entitiesAt(x,y, filter)
  for i=1,len do
    local item = items[i]
    self:collectItem(item)
  end
end

-----------------------------------------------------------
-- update stats
-----------------------------------------------------------
function player:calculateStats()
  local dexbonus = math.floor((self.dex - 10) / 2)
  local strbonus = math.floor((self.str - 10) / 2)
  local ac = 12 + dexbonus
  local tohit = strbonus + self.level

  for k,v in pairs(self.equipment) do

    if v.item ~= nil and v.item.ac then
      ac = ac + v.item.ac
      if v.item.bonus then
        ac = ac + v.item.bonus
      end
    end
  end

  if math.floor(self.level / 5) > 0 then
    tohit = tohit + math.floor(self.level / 5)
  end

  self.ac = ac
  self.tohit = tohit
  self.damagebonus = math.max(strbonus, 0)

  self.hp = math.min(self.hp, self.maxhp)
end

function player:gainXP(amt)
  self.xp = self.xp + amt
  if self.xp >= 10 * self.level then
    self:levelUp()
  end
end

function player:levelUp()
  self.level = self.level + 1
  self.xp = 0
  local hproll = 6
  self.maxhp = self.maxhp + hproll
  self.hp = self.hp + hproll


  self.str = self.str + 1
  self.dex = self.dex + 1
  self.mind = self.mind + 1


  self:calculateStats()
  Signal.emit('message', 'You advanced to level ' .. self.level .. '!')
end

-----------------------------------------------------------
-- Items and map features
-----------------------------------------------------------
function player:interact()
  local x, y = self.x, self.y

  local entities, len = Game:entitiesAt(x,y)
  -- items
  for i=1,len do
    local e = entities[i]
    if e.isItem then
      self:collectItem(e)
    elseif e.mapfeaturetype == 'door' and e.open then
      self:closeDoor(e)
    elseif e.mapfeaturetype == 'stairsdown' then
      self:goDownStairs()
    end
  end

end

function player:collectItem(item)
  -- collect gold and keys
  local bindings = {
    gold = function() self:collectGold(item) end,
    key = function() self:collectKey(item) end
  }
  local action = bindings[item.itemtype]

  if type(action) == 'function' then
    action()
  elseif #self.inventory <= self.maxInventorySize then
    self:addItemToInventory(item)
  else
    Signal.emit('message', 'Your inventory is full!')
  end
end

function player:collectGold(goldItem)
  local value = goldItem.value
  local qty = value == 1 and 'gold piece.' or 'gold pieces.'
  local str = 'You picked up ${v} ${q} ' % {v = value, q = qty}
  self.gold = self.gold + value
  Signal.emit('message', str)
  Signal.emit('addEntity', floatingText({
    text = '+' .. value,
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize,
    color = colors.yellow
  }))
  goldItem._remove = true
end

function player:collectKey(key)
  local value = key.value
  local str = 'You picked up '

  if value == 1 then
    str = str .. 'a key.'
  else
    str = str .. value .. ' keys.'
  end
  self.keys = self.keys + value
  Signal.emit('message', str)
  Signal.emit('addEntity', floatingText({
    text = '+' .. value,
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize,
    color = colors.white
  }))
  key._remove = true
end

function player:addItemToInventory(item)
  local inv = self.inventory

  if item.isStackable then
    local found = false
    for i=1, #inv do
      if inv[i].name == item.name then
        found = true
        inv[i].quantity = inv[i].quantity + 1
        break
      end
    end
    if not found then
      local info = item:getInventoryInfo()
      info.quantity = 1
      table.insert(self.inventory, info)
    end
  else
    table.insert(self.inventory, item:getInventoryInfo())
  end

  local article = ''
  if item.name:match('^[aeiouAEIOU]') then
    article = 'an'
  else
    article = 'a'
  end
  local str = 'You picked up ${a} ${n}.' % {a = article, n = item.name}
  Signal.emit('message', str)
  Signal.emit('addEntity', floatingText({
    text = item.name,
    x = self.drawX + self.w / 2,
    y = self.drawY - settings.tilesize,
    color = colors.white,
  }))

  item._remove = true
end

function player:useItem(index)
  local item = self.inventory[index]
  local config = itemlist[item.itemtype]
  if config.isConsumable then
    local remove = config.onUse(item, self) or false

    if remove then
      item.quantity = item.quantity - 1
      if item.quantity <= 0 then
        self:removeItemFromInventory(item)
      end
    end
  end
end

function player:tossItem(index)
  local item = self.inventory[index]
  local str = 'You tossed the ${n}.' % {n = item.name}
  Signal.emit('message', str)
  self:removeItemFromInventory(item)
end

function player:equipItem(index)
  local item = self.inventory[index]
  local config = itemlist[item.itemtype]
  local slot = config.slot
  local currentlyEquipped = self.equipment[slot].item

  if currentlyEquipped ~= nil then
    self:unequipItem(currentlyEquipped, true)
  end

  self.equipment[slot].item = item
  self:removeItemFromInventory(item)

  local str = 'You equipped the ${n}.' % {n = item.name}
  Signal.emit('message', str)
  self:calculateStats()
end

function player:unequipItem(item, silent)
  local silent = silent or false
  local config = itemlist[item.itemtype]
  local slot = config.slot

  table.insert(self.inventory, item)
  self.equipment[slot].item = nil
  if not silent then
    local str = 'You unequipped the ${n}.' % {n = item.name}
    Signal.emit('message', str)
  end
  self:calculateStats()

end

function player:removeItemFromInventory(item)
  local index = table.indexOf(self.inventory, item)
  table.remove(self.inventory, index)
end

-----------------------------------------------------------
-- Doors
-----------------------------------------------------------
function player:openDoor(door)
  door:openDoor()
  local str = 'You open the door.'
  Signal.emit('message', str)
end

function player:closeDoor(door)
  door:closeDoor()
  local str = 'You close the door.'
  Signal.emit('message', str)
end

-----------------------------------------------------------
-- Stairs
-----------------------------------------------------------
function player:goDownStairs()
  Signal.emit('descend')
end


-----------------------------------------------------------
-- Saving
-----------------------------------------------------------
function player:save()
  return {
    type = 'Player',
    x = self.x,
    y = self.y,
    args = {
      hp = self.hp,
      maxhp = self.maxhp,
      str = self.str,
      dex = self.dex,
      mind = self.mind,
      gold = self.gold,
      keys = self.keys,
      ac = self.ac,
      level = self.level,
      xp = self.xp,
      statuses = self.statuses,
      quadname = self.quadname
    },
    inventory = self.inventory,
    equipment = self.equipment
  }
end

return player
