local colors = require 'modules.setup.colors'
local settings = require 'modules.setup.settings'
local utils = require 'modules.setup.utils'
local itemlist = require 'modules.items.itemlist'
local menu = require 'modules.ui.menu'

local defaults = {
  isActive = false,
}
local equipmentmenu = Class {__includes = {
  menu,
  defaults
}}

-----------------------------------------------------------
-- Init
-----------------------------------------------------------
function equipmentmenu:init(parent)
  self.parent = parent
  self.x = 0
  self.y = settings.tilesize
  self.maxLines = 12
  self.bg = colors.darkgreen
  self.options = {
    'Cancel',
  }
  self:onSelect(function(option)
    local bindings = {
      ['Cancel'] = function() self:deactivate() end,
    }
    local action = bindings[option]
    if type(action) == 'function' then
      action()
    else
      --self:openSubmenu()
    end
  end)
end

-----------------------------------------------------------
-- Activation
-----------------------------------------------------------
function equipmentmenu:activate()
  self:onOpen()
  self.selection = 0
  self.isActive = true
  self.submenu = nil
end

function equipmentmenu:deactivate()
  local parent = self.parent
  self.isActive = false
  parent.menu.isActive = true
  parent.currentmenu = parent.menu
end

function equipmentmenu:toggle()
  if self.isActive then
    self:deactivate()
  else
    self:activate()
  end
end

function equipmentmenu:onOpen()
  local player = self.parent.player
  player:calculateStats()
  local equipment = self.parent.player.equipment
  self.equipment = equipment
  self.w, self.h = 0, 0

  local arr = {
    'Cancel'
  }
  for k,v in pairs(equipment) do
    local slot = v
    table.insert(arr, 1, k)
  end
  self.options = arr
  if self:selected() + 1 > #self.options then
    self.selection = #self.options-1
  end
  self:updateDimensions()
end

function equipmentmenu:handleAction(action)
  --self:closeSubmenu()

  local selfinv = self.inventory
  local selected = self:selected() + 1

  if action == 'toss' then
    self.parent.player:tossItem(selected)
  elseif action == 'equip' then
    self.parent.player:equipItem(selected)
  elseif action == 'info' then
    -- TODO
  end
  self:onOpen()
end

function equipmentmenu:updateDimensions()
  local opts = self.options
  self.h = #opts * settings.tilesize
  local w = QualeFont:getWidth('Cancel')
  for i=1, #opts do
    local str = '> '
    local slotname = opts[i]

    if self.equipment[slotname] then
      str = str .. self.equipment[slotname].name
      if self.equipment[slotname].item ~= nil then
        str = str .. self.equipment[slotname].item.name
      end
    end

    local temp = QualeFont:getWidth(str)
    w = math.max(w, temp)
  end
  self.w = w + settings.tilesize * 2
end
-----------------------------------------------------------
-- Submenu
-----------------------------------------------------------
-- function equipmentmenu:openSubmenu()
--   self.submenu = inventorysubmenu(self)
-- end
-- function equipmentmenu:closeSubmenu()
--   self.submenu = nil
-- end
-----------------------------------------------------------
-- Update
-----------------------------------------------------------
function equipmentmenu:update(dt)

end

-----------------------------------------------------------
-- Drawing
-----------------------------------------------------------
function equipmentmenu:draw()
  if self.isActive then
    self:drawBackground()
    self:drawVisibleItems()
    --self:drawSubmenu()
  end
end

function equipmentmenu:drawBackground()
  local x, y, w, h = self.x, self.y, self.w, self.h
  lg.setColor(self.bg)
  lg.rectangle('fill', x, y, w, h)
end

function equipmentmenu:drawVisibleItems()
  local x, y, w, h = self.x, self.y, self.w, self.h
  local opts = self.options
  local equipment = self.equipment
  local lh = settings.tilesize
  local selected = self:selected() + 1
  lg.setColor(self.fg)

  for i=1, #opts do
    local isSelected = i == selected
    local str = ''
    local slotname = opts[i]

    if equipment[slotname] then
      if isSelected then str = '> ' end
      str = str .. equipment[slotname].name .. ': '
      if equipment[slotname].item ~= nil then
        local sw = QualeFont:getWidth(str)
        local item = equipment[slotname].item
        self:drawItem(x + sw, y, item)
      else
        str = str .. '-'
      end
    else
      if isSelected then str = '> ' end
      str = str .. opts[i]
    end
    lg.setColor(colors.white)
    lg.print(str, x, y)
    y = y + lh
  end
end

function equipmentmenu:drawItem(x, y, item)
  local itemtype = item.itemtype
  local config = itemlist[itemtype]
  local fg = config.fg
  local quad = config.quadname
  local str = item.name
  local x,y = x,y
  lg.setColor(fg)
  lg.draw(QualeTiles, QualeQuads[quad], x, y)
  lg.setColor(colors.white)
  lg.print(str, x + settings.tilesize, y)
  lg.setColor(colors.white)
end

-- function equipmentmenu:drawSubmenu()
--   if self.submenu ~= nil then
--     self.submenu:draw()
--   end
-- end
-----------------------------------------------------------
-- menu functions
-----------------------------------------------------------
function equipmentmenu:select()
  if self.submenu ~= nil then
    self.submenu:select()
  else
    local option = self.options[self.selection + 1]
    if self.handlers['select'] then
      self.handlers['select'](option)
    end
  end
end

function equipmentmenu:decrement()
  if self.submenu ~= nil then
    self.submenu:decrement()
  else
    self.selection = (self.selection - 1) % #self.options
  end
end

function equipmentmenu:increment()
  if self.submenu ~= nil then
    self.submenu:increment()
  else
    self.selection = (self.selection + 1) % #self.options
  end
end


return equipmentmenu
