local colors = require 'modules.setup.colors'
local settings = require 'modules.setup.settings'
local utils = require 'modules.setup.utils'
local itemlist = require 'modules.items.itemlist'
local menu = require 'modules.ui.menu'
local inventorysubmenu = require 'modules.ui.menus.inventorysubmenu'
local Text = require 'modules.ui.text'

local defaults = {
  isActive = false,
}
local inventorymenu = Class {__includes = {
  menu,
  defaults
}}

-----------------------------------------------------------
-- Init
-----------------------------------------------------------
function inventorymenu:init(parent)
  self.parent = parent
  self.x = 0
  self.y = settings.tilesize
  self.bg = colors.darkgreen
  self.maxLines = 12
  self.textbox = nil

  self.options = {
    'Cancel',
  }
  self:onSelect(function(option)
    local bindings = {
      ['Cancel'] = function() self:deactivate() end,
    }
    local action = bindings[option]
    if type(action) == 'function' then
      action()
    else
      --self:useItem()
      self:openSubmenu()
    end
  end)
end

-----------------------------------------------------------
-- Activation
-----------------------------------------------------------
function inventorymenu:activate()
  self:onOpen()
  self.selection = 0
  self.displayStart = 1
  self.displayEnd = self.displayStart + self.maxLines
  self.isActive = true
  self.submenu = nil
end

function inventorymenu:deactivate()
  local parent = self.parent
  self.isActive = false
  parent.menu.isActive = true
  parent.currentmenu = parent.menu
end

function inventorymenu:toggle()
  if self.isActive then
    self:deactivate()
  else
    self:activate()
  end
end

function inventorymenu:onOpen()
  local player = self.parent.player
  local inventory = self.parent.player.inventory
  self.inventory = inventory
  self.w, self.h = 0, 0

  local arr = {
    'Cancel'
  }
  for i = 1, #inventory do
    table.insert(arr, 1, inventory[i].name)
  end
  self.options = arr
  if self:selected() + 1 > #self.options then
    self.selection = #self.options-1
  end
  self:updateDimensions()
  self:updateDisplayValues()
end

function inventorymenu:handleAction(action)
  self:closeSubmenu()

  local selfinv = self.inventory
  local selected = self:selected() + 1

  if action == 'use' then
    self.parent.player:useItem(selected)
  elseif action == 'toss' then
    self.parent.player:tossItem(selected)
  elseif action == 'equip' then
    self.parent.player:equipItem(selected)
  elseif action == 'info' then
    self:openTextbox()
  end
  self:onOpen()
end

function inventorymenu:updateDimensions()
  local inv = self.inventory
  self.h = math.min(#inv+1, self.maxLines+1) * settings.tilesize
  local w = QualeFont:getWidth('Cancel')
  for i=1,#inv do
    local item = self.inventory[i]
    local str = item.name
    if item.quantity ~= nil then
      str = str .. ' x' .. item.quantity
    end
    local temp = QualeFont:getWidth(str)
    w = math.max(w, temp)
  end
  self.w = w + settings.tilesize * 2
end
-----------------------------------------------------------
-- Submenu
-----------------------------------------------------------
function inventorymenu:openSubmenu()
  local selected = self:selected() + 1
  local item = self.parent.player.inventory[selected]
  local config = itemlist[item.itemtype]
  local opts = {'Info', 'Toss', 'Cancel'}

  if config.isEquipment then
    table.insert(opts, 1, 'Equip')
  end
  if config.isConsumable then
    table.insert(opts, 1, 'Use')
  end

  self.submenu = inventorysubmenu(self, opts)
end

function inventorymenu:closeSubmenu()
  self.submenu = nil
end

-----------------------------------------------------------
-- textbox
-----------------------------------------------------------
function inventorymenu:openTextbox()
  local selected = self:selected() + 1
  local item = self.parent.player.inventory[selected]
  local config = itemlist[item.itemtype]
  local fg, bg = config.fg, colors.darkgreen

  -- populate text array:
  local lines = {item.name, 'Press [Enter] to close.'}
  local str = item.name
  if config.description then
    table.insert(lines, 2, config.description)
  end
  if item.stats then
    for k, v in pairs(item.stats) do
      table.insert(lines, 2, k:capitalize() .. ': ' .. v)
    end
  end

  if item.bonus then
    local bstr = 'Enchantment: +' .. item.bonus
    if item.bonustype then
      bstr = bstr .. ' ' .. item.bonustype
    end
    table.insert(lines, 2, bstr)
  end

  if item.ac then
    table.insert(lines, 2, 'Armor bonus: ' .. item.ac)
  end

  if item.damage then
    local d,s,m = item.damage[1], item.damage[2], item.damage[3]
    local dstr = d .. 'd' .. s
    if m > 0 then
      dstr = dstr .. '+' .. m
    elseif m < 0 then
      dstr = dstr .. '-' .. m
    end
    table.insert(lines, 2, dstr .. ' Damage')
  end


  -- get dimensions:
  local temp = QualeFont:getWidth(lines[1])
  for i=1, #lines do
    temp = math.max(temp, QualeFont:getWidth(lines[i]))
  end
  -- create text object:
  local str = table.concat(lines, '\n')
  self.textbox = Text({
    text = str,
    centerToScreen = true,
    bgColor = colors.darkgreen
  })
end
-----------------------------------------------------------
-- Update
-----------------------------------------------------------
function inventorymenu:update(dt)

end

-----------------------------------------------------------
-- Drawing
-----------------------------------------------------------
function inventorymenu:draw()
  if self.isActive then
    self:drawBackground()
    self:drawVisibleItems()
    self:drawSubmenu()
    self:drawTextbox()
  end
end

function inventorymenu:drawBackground()
  local x, y, w, h = self.x, self.y, self.w, self.h
  lg.setColor(self.bg)
  lg.rectangle('fill', x, y, w, h)
end

function inventorymenu:drawVisibleItems()
  local x, y, w, h = self.x, self.y, self.w, self.h
  local opts = self.options
  local inventory = self.inventory
  local lh = settings.tilesize
  local selected = self:selected() + 1
  local len = math.min(#opts, self.maxLines)
  lg.setColor(self.fg)

  for i = 1, len do
    local isSelected = i == selected
    if inventory[i] ~= nil then
      local item = inventory[i]
      local qty = item.quantity or nil
      self:drawItem(x, y, item, isSelected, qty)
    else
      local str = ''
      if isSelected then
        str = str .. '> ' .. opts[i]
      else
        str = opts[i]
      end

      lg.print(str, x, y)
    end
    y = y + lh
  end
end

function inventorymenu:drawItem(x, y, item, isSelected, qty)
  local itemtype = item.itemtype
  local config = itemlist[itemtype]
  local fg = config.fg
  local quad = config.quadname
  local str = item.name
  local x,y = x,y
  if isSelected then
    lg.setColor(colors.white)
    lg.print('> ', x, y)
    x = x + settings.tilesize
  end
  if qty ~= nil then
    str = str .. ' x' .. qty
  end
  lg.setColor(fg)
  lg.draw(QualeTiles, QualeQuads[quad], x, y)
  lg.setColor(colors.white)
  lg.print(str, x + settings.tilesize, y)
  lg.setColor(colors.white)
end

function inventorymenu:drawSubmenu()
  if self.submenu ~= nil then
    self.submenu:draw()
  end
end

function inventorymenu:drawTextbox()
  if self.textbox ~= nil then
    self.textbox:draw()
  end
end
-----------------------------------------------------------
-- menu functions
-----------------------------------------------------------
function inventorymenu:select()
  if self.submenu ~= nil then
    self.submenu:select()
  elseif self.textbox ~= nil then
    self.textbox = nil
  else
    local option = self.options[self.selection + 1]
    if self.handlers['select'] then
      self.handlers['select'](option)
    end
  end
end

function inventorymenu:decrement()
  if self.submenu ~= nil then
    self.submenu:decrement()
  elseif self.textbox == nil then
    self.selection = (self.selection - 1) % #self.options
    self:updateDisplayValues()
  end
end

function inventorymenu:increment()
  if self.submenu ~= nil then
    self.submenu:increment()
  elseif self.textbox == nil then
    self.selection = (self.selection + 1) % #self.options
    self:updateDisplayValues()
  end
end

function inventorymenu:updateDisplayValues()
  local sel = self.selection + 1
  if sel == 1 then
    self.displayStart = 1
  elseif sel == #self.options then
    self.displayStart = #self.options - self.maxLines
  elseif sel > self.displayEnd then
    self.displayStart = self.displayStart + 1
  elseif sel < self.displayStart then
    self.displayStart = self.displayStart - 1
  end

  self.displayEnd = math.min(self.displayStart + self.maxLines, #self.options)
end

return inventorymenu
