local settings = require 'modules.setup.settings'
local colors = require 'modules.setup.colors'
local utils = require 'modules.setup.utils'
local messageLog = require 'modules.ui.newmessagelog'
local gamestatemenu = require 'modules.ui.menus.gamestatemenu'
local inventorymenu = require 'modules.ui.menus.inventorymenu'
local equipmentmenu = require 'modules.ui.menus.equipmentmenu'
local statsmenu = require 'modules.ui.menus.statsmenu'

local levelgen = require 'modules.level.levelgen'
local levellist = require 'modules.level.levellist'
local map = require 'modules.level.map'
local mapfeature = require 'modules.level.mapfeature'

local player = require 'modules.player'
local enemy = require 'modules.enemies.enemy'
local item = require 'modules.items.item'
local itemlist = require 'modules.items.itemlist'

local shine = require 'lib.shine'

Game = {}
-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function Game:init()
  self.entities = {}
  self.mapW = 60
  self.mapH = 60
  self.smallCanvas = lg.newCanvas(settings.pxWidth, settings.pxHeight)
  self.camera = Camera(0,0, 1)
  self.cameraSmoother = Camera.smooth.damped(8)

  self.menu = gamestatemenu(self)
  self.inventorymenu = inventorymenu(self)
  self.equipmentmenu = equipmentmenu(self)
  self.statsmenu = statsmenu(self)
  self.currentmenu = nil

  self.log = messageLog()
  self.levelgen = levelgen()

  self:registerSignals()

  -- shine stuff:
  self:setUpShader()

end

function Game:registerSignals()
  Signal.register('addEntity', function(e)
    table.insert(Game.entities, e)
    return e
  end)

  Signal.register('lock', function()
    self:lock()
  end)

  Signal.register('unlock', function()
    self:unlock()
  end)

  Signal.register('updateFOV', function(x,y)
    self:doFOV(x,y)
  end)

  Signal.register('screenshake', function(d)
    self:screenshake(d)
  end)

  Signal.register('resize', function()
    self.map:drawToCache()
    self:setUpShader()
  end)

  Signal.register('descend', function()
    self:descend()
  end)
end

function Game:enter(previous, continue)
  if continue == true then
    self:load()
  else
    self:reset()
  end
end

function Game:setUpShader()
  local crt = shine.crt()
  crt.parameters = {x = 0.02, y = 0.02}
  local scanlines = shine.scanlines()
  scanlines.parameters = {pixel_size = 3, line_height = 0.35}
  local boxblur = shine.boxblur()
  boxblur.parameters = {radius=.3}
  self.post_effect = boxblur:chain(scanlines):chain(crt)
end
-----------------------------------------------------------
-- testing new newGame/loadGame functions
-----------------------------------------------------------
function Game:newResetFunction()
  self._locked = false
  self.map = nil
  self.depth = 1
  self.entities = {}
end

function Game:newGame()
  self:newResetFunction()

  local e, m = self.levelgen:makeLevel(self.depth)
  self.entities = e
  self.map = m

  self.camera.lookAt(self.player.x, self.player.y)
  self:postSetup()
end

function Game:loadGame(savedata)
  self:newResetFunction()
  self.depth = savedata.depth
  self.map = map(savedata.map)
  self.entities = self:loadEntities(savedata.entities)

  -- random seed:
  if savedata.randomseedlow ~= nil and savedata.randomseedhigh ~= nil then
    love.math.setRandomSeed(savedata.randomseedlow, savedata.randomseedhigh)
  end

  self.camera:lookAt(savedata.camerapos.x, savedata.camerapos.y)
  self:postSetup()
end

function Game:loadEntities(entities)
  local arr = {}

  for _,e in ipairs(entities) do
    local ent
    if e.type == 'Player' then
      ent = player(e.x, e.y, e.args)
      self.player = ent
      self.player.inventory = utils.clone(e.inventory)
      self.player.equipment = utils.clone(e.equipment)
    elseif e.type == 'Enemy' then
      ent = enemy(e.x, e.y, e.args)
    elseif e.type == 'Item' then
      ent = item(e.x, e.y, e.args)
    elseif e.type == 'Mapfeature' then
      ent = mapfeature(e.x, e.y, e.args)
    end
    table.insert(arr, ent)
  end

  return arr
end

function Game:postSetup()
  self:doFOV(self.player.x, self.player.y)
  self.randomseedlow, self.randomseedhigh = love.math.getRandomSeed()
end
-----------------------------------------------------------
-- end testing
-----------------------------------------------------------
function Game:reset(savedata)
  self._locked = false
  self.map = nil
  self.depth = 1
  self.entities = {}

  if savedata then
    self:loadGame(savedata)
  else
    self.map, self.entities = self.levelgen:makeLevel(self.depth)

    -- spawn player at stairs position
    local x, y
    for i=1, #self.entities do
      local e = self.entities[i]
      if e.mapfeaturetype == 'stairsup' then
        x,y = e.x, e.y
        break
      end
    end
    self.player = player(x, y)
    Signal.emit('addEntity', self.player)

    -- Generated new map message:
    Signal.emit('message', 'You have entered the dungeon.')
  end

  -- set up the camera
  local cx, cy
  if savedata and savedata.camera then
    cx, cy = savedata.camera.x, savedata.camera.y
  else
    cx, cy = self.player.drawX, self.player.drawY
  end
  self.camera:lookAt(cx, cy)

  self:doFOV(self.player.x, self.player.y)
  self.randomseedlow, self.randomseedhigh = love.math.getRandomSeed()
end

function Game:descend()
  local d = self.depth + 1
  local windepth = #levellist + 1
  if d == windepth then
    Gamestate.switch(WinGame)
  elseif levellist[d] then
    self.depth = d
    local p = self.player
    local m, e = self.levelgen:makeLevel(d)
    self.map = m
    self.entities = e

    Signal.emit('addEntity', self.player)
    -- spawn player at stairs position
    local x, y
    for i=1, #self.entities do
      local e = self.entities[i]
      if e.mapfeaturetype == 'stairsup' then
        x,y = e.x, e.y
        break
      end
    end
    self.player:teleportTo(x,y)
    self.camera:lookAt(self.player.x, self.player.y)


    local str = 'You descend to level ' .. d ..'.'
    Signal.emit('message', str)
  end
end

-----------------------------------------------------------
-- Update functions
-----------------------------------------------------------
function Game:update(dt)
  for _,item in ipairs(self.entities) do
    if item['update'] then item:update(dt) end
  end
  self:cameraFollowPlayer()
  self.log:update()

  Game:removeItems()
end

function Game:removeItems()
  for i=#self.entities,1,-1 do
    local item = self.entities[i]
    if item._remove then
      table.remove(self.entities, i)
    end
  end
end

-----------------------------------------------------------
-- Drawing functions
-----------------------------------------------------------
function Game:draw()
  lg.setCanvas(self.smallCanvas)
  lg.clear(colors.black)
  lg.setColor({255,255,255})
  lg.push()

  -- draw world stuff
  self.camera:attach()
  self.map:draw()
  self:drawEntities()
  self.camera:detach()

  self.log:draw()
  self:drawPlayerStats()
  if self.currentmenu ~= nil then self.currentmenu:draw() end
  lg.pop()
  lg.setCanvas()

  if USERSETTINGS.screenEffects then
    self.post_effect:draw(function()
      lg.draw(self.smallCanvas, 0, 0, 0, GscaleX, GscaleY)
    end)
  else
    lg.draw(self.smallCanvas, 0, 0, 0, GscaleX, GscaleY)
  end
end

function Game:drawEntities()
  table.sort(self.entities, function(a,b) return a.zIndex < b.zIndex end)
  for i=1,#self.entities do
    local item = self.entities[i]
    if item['draw'] then item:draw() end
  end
end

function Game:screenshake(duration)
  -- shake the camera for one second
  local orig_x, orig_y = self.camera:position()
  local d = duration or .1
  Timer.during(d, function()
      self.camera:lookAt(orig_x + math.random(-2,2), orig_y + math.random(-2,2))
  end, function()
      -- reset camera position
      self.camera:lookAt(orig_x, orig_y)
  end)
end

function Game:drawPlayerStats()
  local p = self.player
  local x,y,w,h = 0, 0, settings.pxWidth, settings.tilesize
  if p == nil then return end

  -- draw background
  lg.setColor(colors.darkgreen)
  lg.rectangle('fill', x, y, w, h)
  -- draw hp
  lg.setColor(colors.red)
  lg.draw(QualeTiles, QualeQuads['heart'], x, y)

  lg.setColor(colors.white)
  local str = '  ' .. p.hp .. '/' .. p.maxhp .. ' '
  lg.print(str, x, y)

  -- draw keys
  x = x + QualeFont:getWidth(str)
  lg.setColor(colors.gray)
  lg.draw(QualeTiles, QualeQuads['key'], x, y)
  str = '  ' .. p.keys .. ' '
  lg.print(str, x, y)

  -- draw gold
  x = x + QualeFont:getWidth(str)
  lg.setColor(colors.yellow)
  lg.draw(QualeTiles, QualeQuads['gold'], x, y)
  str =  '  ' .. p.gold .. ' '
  lg.print(str, x, y)

  -- draw depth
  x = x + QualeFont:getWidth(str)
  lg.setColor(colors.gray)
  lg.draw(QualeTiles, QualeQuads['stairsdown'], x, y)
  str =  '  ' .. self.depth
  lg.print(str, x, y)

  lg.setColor(colors.white)
end

-----------------------------------------------------------
-- Input functions
-----------------------------------------------------------
function Game:keypressed(key, code)
  if self.currentmenu ~= nil then
    self:handleMenuInput(key, code)
  else
    self:handlePlayerInput(key, code)
  end
end

function Game:handleMenuInput(key, code)
  local menu = self.currentmenu

  local keys = {
    w =           'decrement',
    up =          'decrement',
    s =           'increment',
    down =        'increment',
    ['escape'] =  'deactivate',
    ['return'] =  'select'
  }

  local bindings = {
    decrement = function() menu:decrement() end,
    increment = function() menu:increment() end,
    deactivate = function() menu:deactivate() end,
    select = function() menu:select() end
  }

  local func = keys[key]
  local action = bindings[func]
  if action then action() end
end

function Game:handlePlayerInput(key, code)
  local keys = {
    a =     'left',
    left =  'left',
    d =     'right',
    right = 'right',
    w =     'up',
    up =    'up',
    s =     'down',
    down =  'down',
    space = 'interact',
    r =     'reset',
    ['1'] = 'save',
    ['2'] = 'load',
    ['3'] = 'postdraw',
    ['escape'] =     'menu',
    ['return'] = 'gameover'
  }

  local bindings = {
    left = function() self:movePlayer(-1, 0) end,
    right = function() self:movePlayer(1, 0) end,
    up = function() self:movePlayer(0, -1) end,
    down = function() self:movePlayer(0, 1) end,
    interact = function() self:playerInteract() end,
    reset = function() self:reset() end,
    save = function() self:save() end,
    load = function() self:load() end,
    postdraw = function()
      USERSETTINGS.screenEffects = not USERSETTINGS.screenEffects
    end,
    menu = function()
      self.currentmenu = self.menu
      self.menu:activate()
    end,
    gameover = function()
      if self.player._dead then
        self:deleteSaveFile()
        self:reset()
      end
    end
  }
  local func = keys[key]
  local action = bindings[func]
  if action then action() end
end

function Game:mousepressed(x, y, button, istouch)
  -- to do
end

-----------------------------------------------------------
-- Acting functions
-----------------------------------------------------------
function Game:lock()
  self._locked = true
end

function Game:unlock()
  if self._locked == false then return end
  self._locked = false

  for i=1,#self.entities do
    local item = self.entities[i]
    if item['act'] then item:act() end
  end
end

-----------------------------------------------------------
-- Player functions
-----------------------------------------------------------
function Game:movePlayer(dx, dy)
  local player = self.player
  local tiles = self.map.tiles
  local filter = function(e) return e.blocksMovement == true end

  if player and (player._dead == false) and (self._locked) == false then
    local x,y = player.x + dx, player.y + dy
    local entities, len = self:entitiesAt(x,y, filter)
    -- check for tile collision:
    if tiles[y] and tiles[y][x] == 1 then
      local str = 'You cannot go that way.'
      Signal.emit('message', str)
    elseif len > 0 then
      self.player:blockedPath(x, y, entities, len)
    else
      self.player:moveTo(x, y)
    end
  end
end

function Game:playerInteract()
  -- it may seem clunky to have a playerInteract function instead of just calling
  -- self.player:interact, but I feel like this could avoid a weird situation
  -- where the player interacts with something while the engine is supposed to
  -- be locked (like during movement tweens) that could cause problems
  local player = self.player
  if player and (player._dead == false) and (self._locked) == false then
    player:interact()
  end
end

function Game:cameraFollowPlayer()
  local x, y = self.player.drawX, self.player.drawY
  local border = 96
  local xmin, xmax = border, settings.pxWidth - border
  local ymin, ymax = border, settings.pxHeight - border
  local smoother = self.cameraSmoother
  self.camera:lockWindow(x, y, xmin, xmax, ymin, ymax, smoother)
end

-----------------------------------------------------------
-- Map functions
-----------------------------------------------------------
function Game:spaceIsFree(x, y)
  local tiles = self.map.tiles
  local entities, len = self:entitiesAt(x,y)
  if (tiles[y] and tiles[y][x] == 0) and (len == 0) then
    return true
  else
    return false
  end
end

function Game:doFOV(x,y)
  local blockingEntities, len = self:getEntities(function(e)
    return e.blocksVision == true
  end)

  self.map:doFOV(x,y, blockingEntities)
  for _,item in ipairs(self.entities) do
    if item['updateVisibility'] then
      item:updateVisibility(self.map.visible)
    end
  end
end

-----------------------------------------------------------
-- Entity functions
-----------------------------------------------------------
function Game:moveEntityTo(entity, x, y)
  local tiles = self.map.tiles
  local filter = function(e)
    return e.blocksMovement or e.isPlayer
  end
  local entities, len = self:entitiesAt(x,y, filter)
  -- check for tile collision:
  if tiles[y] and tiles[y][x] == 1 then
    return false
  elseif len > 0 then
    entity:blockedPath(x, y, entities, len)
  else
    entity:moveTo(x, y)
  end
end

function Game:entitiesAt(x,y, filter)
  local list = {}
  local f = filter or function(e) return true end

  for _,item in ipairs(self.entities) do
    if item.x == x and item.y == y and f(item) then
      table.insert(list, item)
    end
  end
  return list, #list
end

function Game:getEntities(filter)
  local list = {}
  local f = filter or function(e) return true end
  for _,item in ipairs(self.entities) do
    if f(item) then
      table.insert(list, item)
    end
  end
  return list, #list
end

-----------------------------------------------------------
-- Save and load functions
-----------------------------------------------------------
function Game:save()
  local savegame = {}
  savegame.depth = self.depth
  savegame.map = self.map:save()
  savegame.entities = {}
  savegame.randomseedlow, savegame.randomseedhigh = love.math.getRandomSeed()
  savegame.camerapos = {
    x = self.camera.x,
    y = self.camera.y
  }

  for _,item in ipairs(self.entities) do
    if item['save'] then
      local data = item:save()
      table.insert(savegame.entities, data)
    end
  end
  lf.write(settings.savefilename, JSON.encode(savegame))
  Signal.emit('message', 'Game saved.')
end

function Game:load()
  if lf.exists(settings.savefilename) then
    local str = lf.read(settings.savefilename)
    local savegame = JSON.decode(str)

    self:reset(savegame)
    Signal.emit('message', 'Game loaded.')
  else
    Signal.emit('message', 'Save file does not exist. :(')
    return false
  end
end

function Game:deleteSaveFile()
  if lf.exists(settings.savefilename) then
    lf.remove(settings.savefilename)
  end
end
